#include <iostream.h>
#include <vector.h>
#include <gen_interval_tree.h>

/* type of an interval */
struct Interval
{
  double left, right;
  Interval(double l, double r): left(l), right(r)
  {}
  Interval()
  {}
};

/* input operator for an interval: */
istream &operator>> (istream &in, Interval& interval)
{
  in >> interval.left >> interval.right;
  return in;
}

/* output operator for an interval: */
ostream &operator<< (ostream &out, Interval const &interval)
{
  out << "(" << interval.left << ", " << interval.right << ")";
  return out;
}

/* data accessors for the left and right endpoints of the intervals */
struct LeftDA { typedef double value_type; };
struct RightDA { typedef double value_type; };

/* the corresponding data accessing functions */
template <class It>
inline double get(LeftDA da, It const &it) { return (*it).left; }
template <class It>
inline double get(RightDA da, It const &it) { return (*it).right; }

main()
{
  cout << endl << "         Testing the interval tree" << endl << endl;

  /* type definitions */
  typedef vector<Interval> Interval_vector;
  typedef typename Interval_vector::iterator Interval_vector_it;
  typedef GenIntervalTree<Interval_vector_it, LeftDA, RightDA> Interval_tree;
  typedef typename Interval_tree::Interval_position ID;
  typedef typename Interval_tree::Iterator TreeIt;

  /* read the input intervals */
  Interval_vector interval_ls;
  Interval interval;
  cout << "Enter the endpoints of the intervals: " << endl <<
    "(enter an interval of length zero to mark the end of the input" <<
    " intervals)"<< endl;
  while (cin >> interval && (interval.left != interval.right))
    interval_ls.push_back(interval);

  /* output the read intervals */
  cout << "  Interval No.    " << "Interval" << endl;
  int i = 0;
  Interval_vector_it intervals_end = interval_ls.end();
  for (Interval_vector_it intervals_beg = interval_ls.begin();
       intervals_beg != intervals_end; intervals_beg++)
    cout << "      " << i++ << "           " << *intervals_beg << endl;
  cout << endl;

  /* declare an object of type interval tree */
  Interval_tree interval_tree(interval_ls.begin(), interval_ls.end());
  ID *del_info = new ID[interval_ls.size()];

  /* update on and report from the interval tree */
  cout << "i-insert, d-delete, r-report and number of interval:" << endl;
  char c;
  while (cin >> c)
  {
    cin >> i;
    switch (c)
      {
      case 'i':
        {
          del_info[i] = interval_tree.insert(&interval_ls[i]);
        }
        break;
      case 'd': 
        {
          interval_tree.erase(del_info[i]);
        }
        break;
      case 'r':
        {
          TreeIt end = interval_tree.end();
          TreeIt begin = interval_tree.begin(&interval_ls[i]);
          for(; !(begin == end); ++begin)
            cout << *(*begin) << endl;
        }
        break;
      }
    }
    delete[] del_info;
  }
