#ifndef GEN_INTER_TREE
#define GEN_INTER_TREE

#if (__SUNPRO_CC)
#define typename
#endif

#include <set.h>
#include <multimap.h>
#include <list.h>

// class GenIntervalTree_user_compare expects compare function object from user
template <class ForwardIt, class LeftDA, class RightDA, class Compare>
class GenIntervalTree_user_compare
{
public:   
    // position of an interval in interval tree
    class Interval_position_info;
    typedef typename list<Interval_position_info*>::iterator Interval_position;
private:
    typedef typename LeftDA::value_type value_type;
    typedef pair<ForwardIt, Interval_position> Data_pair;
    typedef multimap<value_type, Data_pair, less<value_type> >
    Endpoint_list;
    typedef typename Endpoint_list::iterator Endpoint_list_it;
    typedef typename Endpoint_list::value_type Value_type_and_data_pair;
    typedef set<value_type, less<value_type> > Endpoint_set;
    typedef typename Endpoint_set::iterator Endpoint_set_it;

    // type of a node in interval tree
    struct Tnode;
    friend struct Tnode;
    struct Tnode
    {
	Tnode(): left(NULL), right(NULL)
	{}
	// bool active;
	value_type discriminant;
	Tnode* left, *right;
	Endpoint_list left_endpoints, right_endpoints;
    };

    // position of an interval in interval tree
    class Interval_position_info;
    friend class Interval_position_info;
    class Interval_position_info
    {
    public:
	Interval_position_info()
	{}
	Interval_position_info(Tnode* nd)
	    : node(nd)
	{}
	~Interval_position_info()
	{
	    node->left_endpoints.erase(left_endpoint);
	    node->right_endpoints.erase(right_endpoint);
	}
	void set_endpoints(Endpoint_list_it l_it, Endpoint_list_it r_it)
	{
	    left_endpoint = l_it;
	    right_endpoint = r_it;
	}
    private:
	Tnode* node;
	Endpoint_list_it left_endpoint, right_endpoint;
    };

public:
    // class constructor
    GenIntervalTree_user_compare( ForwardIt beg, ForwardIt end, 
				  LeftDA lda = LeftDA(), 
				  RightDA rda = RightDA(), 
				  Compare com = Compare() ) 
	: my_leftDA(lda), my_rightDA(rda), my_compare(com)
    {
	make_skeleton(beg, end);
    }
  
    // class destructor
    virtual ~GenIntervalTree_user_compare()
    {
	Interval_position current = interval_position_infos.begin();
	for ( ; current != interval_position_infos.end(); current++)
	    if (*current) delete *current;
	if (number_of_leaves > 0)
	    delete[] root;
    }

    // the iterator used to iterate over the solution found
    class Iterator;
    friend class Iterator;
    class Iterator
    {
	enum direction {left, right};
    public:
	Iterator()
	    : valid(false)
	{}
	
	Iterator(ForwardIt interval,
		 LeftDA lda, RightDA rda,
		 Tnode* root,
		 Compare com)
	    : my_query_interval(interval), current_node(root), my_cmp(com)
	{
	    my_left_end = get(lda, interval);
	    my_right_end = get(rda, interval);
	    
	    find_next_node();
	    init_iterator();
	}
	bool operator==(const Iterator &it)
	{
	    if (!valid)
		return (valid == it.valid);
	    if (it.valid)
	    {
		if ((my_direction == left) && (it.my_direction == left))
		    return (my_left_it == it.my_left_it);
		if ((my_direction == right) && (it.my_direction == right))
		    return (my_right_it == it.my_right_it);
	    }
	    return false;
	}
    
	bool operator!=(const Iterator &it)
	{
	    return !(*this == it);
	}
	ForwardIt operator*()
	{
	    if (my_direction == left) return (*my_left_it).second.first;
	    return (*my_right_it).second.first;
	}
    
	Interval_position get_position()
	{
	    if (my_direction == left) return (*my_left_it).second.second;
	    return (*my_right_it).second.second;
	}

	Iterator& operator++()
	{
	    if (my_direction == left)
	    {
		my_left_it++;
		if ((my_left_it == current_node->left_endpoints.end()) ||
		    (!my_cmp((*my_left_it).first, my_right_end)))
		{
		    current_node = current_node->left;
		    find_next_node();
		    init_iterator();
		}
	    }
	    else
	    {
		my_right_it++;
		if ((my_right_it == current_node->right_endpoints.rend()) ||
		    (!my_cmp(my_left_end, (*my_right_it).first)))
		{
		    current_node = current_node->right;
		    find_next_node();
		    init_iterator();
		}
	    }	    
	    increment_on_identity();
	    return *this;
	}
    private:
	ForwardIt my_query_interval;
	bool valid;
	value_type my_left_end, my_right_end;
	Tnode *current_node;
	Endpoint_list::iterator my_left_it;
	Endpoint_list::reverse_iterator my_right_it;
	direction my_direction;
	Compare my_cmp;
	list<Tnode*> node_stack;

	void find_next_node()
	{
	    if (current_node == NULL)
	    {
		if (node_stack.empty()) { valid = false; return;}
		else
		{
		    current_node = node_stack.front();
		    node_stack.pop_front();
		}
	    }
	    if (my_left_end < current_node->discriminant)
	    {
		my_direction = left;
		if ((current_node->discriminant <= my_right_end) &&
		    (current_node->right != NULL))
		    node_stack.push_front(current_node->right);
		if (!(current_node->left_endpoints.empty()) &&
		    (my_cmp((*(current_node->left_endpoints.begin())).first,
			    my_right_end)))
		    valid = true;
		else
		{
		    current_node = current_node->left;
		    find_next_node();
		}
	    }
	    else
	    {
		my_direction = right;
		if (!(current_node->right_endpoints.empty()) &&
		    (my_cmp(my_left_end,
			    (*(current_node->right_endpoints.rbegin())).first)))
		    valid = true;
		else
		{
		    current_node = current_node->right;
		    find_next_node();
		}
	    }
	}
	void init_iterator()
	{
	    if (valid)
	    {
		if (my_direction == left)
		    my_left_it = current_node->left_endpoints.begin();
		else my_right_it = current_node->right_endpoints.rbegin();
	    }
	    increment_on_identity();
	}
    
	// increment iterator if current interval is identical to query interval
	void increment_on_identity()
	{
	    if (  valid &&
		  ( ((my_direction == left) &&
		     ((*my_left_it).second.first == my_query_interval)) ||
		    ((my_direction == right) &&
		     ((*my_right_it).second.first == my_query_interval)) )  )
		++(*this);
	}
    };

    // iterators used to iterate over the solution found
    Iterator begin(ForwardIt const &interval)
    {
	return Iterator(interval, my_leftDA, my_rightDA, root, my_compare);
    }  
    Iterator end()
    {
	return Iterator();
    }

    // insert an interval in the interval tree
    Interval_position insert(ForwardIt interval)
    {
	value_type left_end = get(my_leftDA, interval);
	value_type right_end = get(my_rightDA, interval);
	
	bool found = false;
	Tnode* current_node = root;
	while (!found)
	{
	    if (right_end < current_node->discriminant)
		current_node = current_node->left;
	    else if (left_end > current_node->discriminant)
		current_node = current_node->right;
	    else found = true;
	}
    
	Interval_position_info* position_info =
	    new Interval_position_info(current_node);
	Interval_position position_it =
	    interval_position_infos.insert(interval_position_infos.end(),
					   position_info);
	
	// insert the left endpoint in the list of left endpoints
	Endpoint_list_it it_left =
	    current_node->left_endpoints.insert(
		Value_type_and_data_pair(left_end, 
					 Data_pair(interval, position_it)) );
	// insert the right endpoint in the list of right endpoints
	Endpoint_list_it it_right =
	    current_node->right_endpoints.insert(
		 Value_type_and_data_pair(right_end,
					  Data_pair(interval, position_it)) ); 
	position_info->set_endpoints(it_left, it_right);
	return  position_it;
    }

    bool is_erased(Interval_position const &interval_pos) const
    {
	return *interval_pos == NULL;
    }
  
    // delete interval from interval tree
    bool erase(Interval_position& interval_pos)
    {
	bool erased_now = false;
	
	if ( !is_erased(interval_pos) )
	{
	    delete *interval_pos;
	    *interval_pos = NULL;
	    erased_now = true;
	}
	return erased_now;
    }  
    void erase(Iterator& it)
    {
	erase(it.get_position());
    }
private:
    Endpoint_set endpoint_set;
    Endpoint_set_it endpoint_set_it;
    Tnode* root;
    int number_of_leaves;
    list<Interval_position_info*> interval_position_infos;
    LeftDA my_leftDA;
    RightDA my_rightDA;
    Compare my_compare;

    // make the inteval tree skeleton
    void make_skeleton(ForwardIt const &beg, ForwardIt const &end)
    {  
	for (ForwardIt my_current = beg; my_current != end; ++my_current)
	{
	    endpoint_set.insert(get(my_leftDA, my_current));
	    endpoint_set.insert(get(my_rightDA, my_current));
	}
	number_of_leaves = endpoint_set.size();
	
	if (number_of_leaves > 0)
	{
	    root = new Tnode[number_of_leaves-1];
	    Tnode *last_parent = root + ((number_of_leaves-1)/2),
		*child = root + 1, *last_child = root + (number_of_leaves-1);
	    for (Tnode* parent = root; parent < last_parent; parent++)
	    {
		parent->left = child;
		child++;
		if (child < last_child)
		    parent->right = child++;
	    }
	    // set the discriminant for the nodes
	    endpoint_set_it = endpoint_set.begin();
	    assign_discriminant(root);
	}
    }

    // assign discriminant to the node
    void assign_discriminant(Tnode* node)
    {
	if (node == NULL) return;
	
	assign_discriminant(node->left);
	node->discriminant = (*endpoint_set_it++ + *endpoint_set_it)/2;
    assign_discriminant(node->right);
    }
};

// compare function object used by class GenIntervalTree
template <class DA>
class Compare
{
    typedef typename DA::value_type value_type;
public:
    bool operator()(value_type const v1, value_type const v2)
    {
	return (v1 < v2);
    }
};

// class interval tree
template <class ForwardIt, class LeftDA, class RightDA>
class GenIntervalTree
: public GenIntervalTree_user_compare<ForwardIt, LeftDA, RightDA,
  Compare<LeftDA> >
{
  typedef GenIntervalTree_user_compare<ForwardIt, LeftDA, RightDA,
    Compare<LeftDA> > Interval_tree;
public:
  GenIntervalTree(ForwardIt beg, ForwardIt end,
               LeftDA lda = LeftDA(), RightDA rda = RightDA())
  : Interval_tree(beg, end, lda , rda, Compare<LeftDA>())
  {}
};
#endif
