#ifndef OO_INTER_TREE
#define OO_INTER_TREE

#if (__SUNPRO_CC)
#define typename
#endif

#include <set.h>
#include <multimap.h>
#include <list.h>
#include <iostream.h>

// type of an interval needed by class OOIntervalTree
template <class T>
struct Interval
{
    T left, right;
    Interval(T l, T r): left(l), right(r)
    {}
    Interval()
    {}
};

// class OOIntervalTree 
template <class CoordType, class InfoType>
class OOIntervalTree
{
public:
    // position of an interval in interval tree  
    class Interval_position_info;
    typedef typename list<Interval_position_info*>::iterator Interval_position;
    typedef CoordType value_type;
    typedef list<Interval<CoordType> > Interval_ls;
    typedef typename Interval_ls::iterator Interval_ls_it;
private:  
    typedef pair<Interval_ls_it, InfoType> Data_pair;
    typedef multimap<value_type, Data_pair, less<value_type> >
    Endpoint_list;
    typedef typename Endpoint_list::iterator Endpoint_list_it;
    typedef typename Endpoint_list::value_type Value_type_and_data_pair;
    typedef set<value_type, less<value_type> > Endpoint_set;
    typedef typename Endpoint_set::iterator Endpoint_set_it;
    enum direction {left, right};
  
    // type of a node in interval tree    
    struct Tnode;
    friend struct Tnode;
    struct Tnode
    {
	Tnode(): left(NULL), right(NULL)
	{}
	// bool active;
	value_type discriminant;
	Tnode* left, *right;
	Endpoint_list left_endpoints, right_endpoints;
    };
 
    // position of an interval in interval tree
    class Interval_position_info;
    friend class Interval_position_info;
    class Interval_position_info
    {
    public:
	Interval_position_info()
	{}
	Interval_position_info(Tnode* nd)
	    : node(nd)
	{}
	~Interval_position_info()
	{      
	    node->left_endpoints.erase(left_endpoint);
	    node->right_endpoints.erase(right_endpoint);
	}
	void set_endpoints(Endpoint_list_it l_it, Endpoint_list_it r_it)
	{
	    left_endpoint = l_it;
	    right_endpoint = r_it;
	}
    private:
	Tnode* node;
	Endpoint_list_it left_endpoint, right_endpoint;
    };    
public:
    // class constructor
    OOIntervalTree(Interval_ls&  intervals)
    {
	make_skeleton(intervals.begin(), intervals.end());
    }
    
    // class destructor
    virtual ~OOIntervalTree()
    {
	Interval_position current = interval_position_infos.begin();
	for ( ; current != interval_position_infos.end(); current++)
	    if (*current) delete *current;
	if (number_of_leaves > 0)
	    delete[] root;
    }
  
    // report intersections
    void report(Interval_ls_it interval, list<InfoType>& intersections)
    {
	current_node = root;
	find_next_node((*interval).left, (*interval).right);
	init_iterator(interval);
	while (valid)
	{
	    if (my_direction == left)
		intersections.push_back((*my_left_it).second.second);
	    else intersections.push_back((*my_right_it).second.second);
	    increment_to_next(interval);
	}
    }
    
    // insert interval
    Interval_position insert(Interval_ls_it& interval, InfoType& info)
    {
	value_type left_end = (*interval).left;
	value_type right_end = (*interval).right;
	
	bool found = false;
	Tnode* curr_node = root;
	while (!found)
	{
	    if (right_end < curr_node->discriminant)
		curr_node = curr_node->left;
	    else if (left_end > curr_node->discriminant)
		curr_node = curr_node->right;
	    else found = true;
	}
	
	Interval_position_info* position_info =
	    new Interval_position_info(curr_node);
	Interval_position position_it =
	    interval_position_infos.insert(interval_position_infos.end(),
					   position_info);
	
	// insert the left endpoint in the list of left endpoints
	Endpoint_list_it it_left =
	    curr_node->left_endpoints.insert(
			Value_type_and_data_pair(left_end,
					      Data_pair(interval, info)) );
	// insert the right endpoint in the list of right endpoints
	Endpoint_list_it it_right =
	    curr_node->right_endpoints.insert(
			      Value_type_and_data_pair(right_end,
                                 Data_pair(interval, info)) );
  
	position_info->set_endpoints(it_left, it_right);
	return  position_it;
    }
   
    bool is_erased(Interval_position const &interval_pos) const
    {
	return *interval_pos == NULL;
    }
  
    // delete interval
    bool erase(Interval_position& interval_pos)
    {
	bool erased_now = false;	
	if ( !is_erased(interval_pos) )
	{
	    delete *interval_pos;
	    *interval_pos = NULL;
	    erased_now = true;
	}
	return erased_now;
    }        

private:
    Endpoint_set endpoint_set;
    Endpoint_set_it endpoint_set_it;
    Tnode* root;
    int number_of_leaves;
    list<Interval_position_info*> interval_position_infos;
    list<Tnode*> node_stack;
    Tnode* current_node;
    Endpoint_list::iterator my_left_it;
    Endpoint_list::reverse_iterator my_right_it;
    direction my_direction;
    bool valid;
    
    // make the inteval tree skeleton         
    void make_skeleton(Interval_ls_it const &beg, Interval_ls_it const &end)
    {
	for (Interval_ls_it my_current = beg; my_current != end; ++my_current)
	{
	    endpoint_set.insert((*my_current).left);
	    endpoint_set.insert((*my_current).right);
	}
	number_of_leaves = endpoint_set.size();
	
	if (number_of_leaves > 0)
	{
	    root = new Tnode[number_of_leaves-1];
	    Tnode *last_parent = root + ((number_of_leaves-1)/2),
		*child = root + 1, *last_child = root + (number_of_leaves-1);
	    for (Tnode* parent = root; parent < last_parent; parent++)
	    {
		parent->left = child;
		child++;
		if (child < last_child)
		    parent->right = child++;
	    }
	    // set the discriminant for the nodes
	    endpoint_set_it = endpoint_set.begin();
	    assign_discriminant(root);
	}
    }
    
    // assign discriminant to the node
    void assign_discriminant(Tnode* node)
    {
	if (node == NULL) return;
	
	assign_discriminant(node->left);
	node->discriminant = (*endpoint_set_it++ + *endpoint_set_it)/2;
	assign_discriminant(node->right);
    }

    // find the node in the interval tree containing intervals having 
    // intersections with the given interval
    void find_next_node(value_type my_left_end, value_type my_right_end)
    {
	if (current_node == NULL)
	{
	    if (node_stack.empty()) { valid = false; return;}
	    else
	    {
		current_node = node_stack.front();
		node_stack.pop_front();
	    }
	}
	if (my_left_end < current_node->discriminant)
	{
	    my_direction = left;
	    if ((current_node->discriminant <= my_right_end) &&
		(current_node->right != NULL))
		node_stack.push_front(current_node->right);
	    if (!(current_node->left_endpoints.empty()) &&
		((*(current_node->left_endpoints.begin())).first <= 
		 my_right_end))
		valid = true;
	    else
	    {
		current_node = current_node->left;
		find_next_node(my_left_end, my_right_end);
	    }
	}
	else
	{
	    my_direction = right;
	    if (!(current_node->right_endpoints.empty()) &&
		(my_left_end <= 
		 (*(current_node->right_endpoints.rbegin())).first))
		valid = true;
	    else
	    {
		current_node = current_node->right;
		find_next_node(my_left_end, my_right_end);
	    }
	}
    }
  
    // go to the first interval intersecting the given interval
    void init_iterator(Interval_ls_it interval)
    {
	if (valid)
	{
	    if (my_direction == left)
		my_left_it = current_node->left_endpoints.begin();
	    else my_right_it = current_node->right_endpoints.rbegin();
	}
	increment_on_identity(interval);
    }
  
    void increment_on_identity(Interval_ls_it interval)
    {
	if (  valid &&
	      ( ((my_direction == left) &&
		 ((*my_left_it).second.first == interval)) ||
		((my_direction == right) &&
		 ((*my_right_it).second.first == interval)) )  )
	    increment_to_next(interval);
    }
    
    // go to the next interval intersecting the given interval
    void increment_to_next(Interval_ls_it interval)
    {
	if (my_direction == left)
	{
	    my_left_it++;
	    if ((my_left_it == current_node->left_endpoints.end()) ||
		(!((*my_left_it).first < (*interval).right)))
	    {
		current_node = current_node->left;
		find_next_node((*interval).left, (*interval).right);
		init_iterator(interval);
	    }
	}
	else
	{
	    my_right_it++;
	    if ((my_right_it == current_node->right_endpoints.rend()) ||
		(!((*interval).left < (*my_right_it).first)))
	    {
		current_node = current_node->right;
		find_next_node((*interval).left, (*interval).right);
		init_iterator(interval);
	    }
	}	
	increment_on_identity(interval);
    }   
};  
#endif
