#ifndef OO_REC_INTER
#define OO_REC_INTER

// compiler dependencies
#if (__SUNPRO_CC)
#define typename
#endif

// includes
#include <vector.h>
#include <pair.h>
#include <algo.h>
#include <oo_interval_tree.h>
#include <ugraph.h>
#include <iostream.h>

// type of a rectangle needed by class OORecInters
template <class T> struct Rectangle 
{ 
    T llx, lly, urx, ury; 
    Rectangle()
    {}
    Rectangle(T lx, T ly, T ux, T uy)
	: llx(lx), lly(ly), urx(ux), ury(uy)
    {}
};

// class OORecInters
template <class CoordType>
class OORecInters
{
public:
    typedef CoordType value_type;
    typedef list<Rectangle<CoordType> > Rectangle_ls;
    typedef typename Rectangle_ls::iterator Rectangle_ls_it;
    typedef Ugraph<Rectangle_ls_it, int> Graph;
private:
    typedef typename Graph::Node_it Node;
    typedef OOIntervalTree<CoordType, Node> Sweep_line_status;
    typedef typename Sweep_line_status::Interval_position Sls_position;
    typedef Sweep_line_status::Interval_ls Interval_ls;
    typedef typename Sweep_line_status::Interval_ls_it Interval_ls_it;

    // an event point
    struct Rectangle_point;
    friend struct Rectangle_point;
    struct Rectangle_point
    {
	value_type x_coord;
	Rectangle_ls_it rectangle;
	Sls_position* sls_pos;
	Interval_ls_it interval;
	bool rec_begin;
	Node node;
  
	// constructors
	Rectangle_point() {}
	Rectangle_point(value_type x, Rectangle_ls_it rec, 
			Sls_position* pos_it, Interval_ls_it it, 
			Node n, bool begin = false)
	    : x_coord(x), rectangle(rec), sls_pos(pos_it), rec_begin(begin), 
	      node(n), interval(it)
	{}
    };

    // function object used to sort the eps
    class Rectangle_point_compare;
    friend class Rectangle_point_compare;
    class Rectangle_point_compare
    {
    public:
	bool operator()(Rectangle_point const &rp1, Rectangle_point const &rp2)
	{
	    return (rp1.x_coord < rp2.x_coord);
	}
    };
  
    typedef vector<Rectangle_point> Event_point_schedule;
    typedef typename Event_point_schedule::iterator Eps_it;
public:  
    // class constructor
    OORecInters(Rectangle_ls& rectangles)
    {
	graph = new Graph();
	init_eps(rectangles.begin(), rectangles.end());
	__inplace_stable_sort(eps.begin(), eps.end(),
			      Rectangle_point_compare());
	my_current_ep = eps.begin();
	sls = new Sweep_line_status(intervals);
    }
  
    // class destructor
    virtual ~OORecInters()
    {
	delete[] sls_positions;
	delete sls;
    }

    // run the algorithm
    Graph* run()
    {
	// move iterator to next event point (if not at beginning or end)
	while (my_current_ep != eps.end())
	{
	    // if the current event point corresponds to LEFT edge
	    // => insert rectangle into sls
	    if ((*my_current_ep).rec_begin)
	    {
		list<Node> inters;
		sls->report((*my_current_ep).interval, inters);
		list<Node>::iterator begin = inters.begin(),
		    end = inters.end();
		for ( ; begin != end; begin++)
		    graph->new_edge((*my_current_ep).node, *begin, 0);
		*((*my_current_ep).sls_pos) =
		    sls->insert((*my_current_ep).interval,(*my_current_ep).node);
	    }
	    // if the current event point corresponded to RIGHT edge
	    // => delete rectangle from sls
	    else
		sls->erase(*((*my_current_ep).sls_pos));
	    my_current_ep++;
	}
	return graph;
    }  
    
    // total number of conflicts
    int num_of_conflicts() { return graph->num_of_edges(); }
private:  
    Sweep_line_status* sls;
    Event_point_schedule eps;
    Sls_position* sls_positions;
    Eps_it my_current_ep;
    Interval_ls intervals;
    Graph* graph;
    
    void init_eps(Rectangle_ls_it const &begin, Rectangle_ls_it const &end)
    {
	// number of input rectangles
	int len = lower_distance(begin, end);
	sls_positions = new Sls_position[len];
	Sls_position* pos_it = sls_positions;
	// list of endpoints used to initialize the IntervalTree
	eps.reserve(2*len);
	for (Rectangle_ls_it my_current = begin; my_current != end; 
	     ++my_current)
	{
	    value_type x1 = (*my_current).llx;
	    value_type x2 = (*my_current).urx;
	    Interval_ls_it it = intervals.insert(intervals.end(), 
			Interval<value_type>((*my_current).lly, 
					     (*my_current).ury));  
	    Node n = graph->new_node(my_current);
	    eps.push_back(Rectangle_point(x1, my_current, pos_it, it, n, 
					  true)); 
	    eps.push_back(Rectangle_point(x2, my_current, pos_it++, it, n));
	}
    }
};  
#endif
